/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { TypedEventEmitter } from "@fluid-internal/client-utils";
import { SummaryType } from "@fluidframework/driver-definitions";
import { MessageType, } from "@fluidframework/driver-definitions/internal";
import { mergeStats } from "@fluidframework/runtime-utils/internal";
import { createChildLogger, raiseConnectedEvent, } from "@fluidframework/telemetry-utils/internal";
import { MockContainerRuntimeFactoryForReconnection, MockContainerRuntimeForReconnection, } from "@fluidframework/test-runtime-utils/internal";
import { v4 as uuid } from "uuid";
import { RunWhileConnectedCoordinator, Summarizer, SummaryCollection, SummaryManager, } from "../../summary/index.js";
export class MockContainerRuntimeFactoryForSummarizer extends MockContainerRuntimeFactoryForReconnection {
    createContainerRuntime(dataStoreRuntime, _) {
        const containerRuntime = new MockContainerRuntimeForSummarizer(dataStoreRuntime, this, this.runtimeOptions);
        this.runtimes.add(containerRuntime);
        return containerRuntime;
    }
}
const DefaultSummaryConfiguration = {
    state: "disableHeuristics",
    maxAckWaitTime: 3 * 60 * 1000, // 3 mins.
    maxOpsSinceLastSummary: 7000,
    initialSummarizerDelayMs: 5 * 1000, // 5 secs.
};
export class MockContainerRuntimeForSummarizer extends MockContainerRuntimeForReconnection {
    baseLogger = createChildLogger();
    summarizerClientId;
    summarizer;
    summaryManager;
    connectedState;
    summaryCollection;
    summarizerClientElection;
    constructor(dataStoreRuntime, factory, runtimeOptions = {}) {
        // trackRemoteOps is needed for replaying all ops on creating new ContainerRuntime
        super(dataStoreRuntime, factory, runtimeOptions, { trackRemoteOps: true });
        this.deltaManager.on("op", (message) => {
            this.emit("op", message);
        });
        this.summarizerClientElection = new MockSummarizerClientElection(this.clientId);
        this.connectedState = new MockConnectedState(this.baseLogger, this.clientId);
        this.summaryCollection = new SummaryCollection(this.deltaManager, this.baseLogger);
        const summaryConfiguration = {
            ...DefaultSummaryConfiguration,
            ...runtimeOptions.summaryConfiguration,
        };
        this.summarizer = new Summarizer(this /* summarizerRuntime */, () => summaryConfiguration /* configurationGetter */, this /* ISummarizerInternalsProvider */, {} /* handleContext */, this.summaryCollection, async (runtime) => RunWhileConnectedCoordinator.create(runtime, () => this.deltaManager.active));
        this.summaryManager = new SummaryManager(this.summarizerClientElection, this.connectedState, this.summaryCollection, this.baseLogger, async () => this.summarizer, new MockThrottler());
        this.summaryManager.start();
    }
    nackScheduled = false;
    /** Prepare a SummaryNack to be sent by the server */
    prepareSummaryNack() {
        this.nackScheduled = true;
    }
    /** Call on the Summarizer object to summarize */
    async summarize() {
        const result = this.summarizer.summarizeOnDemand({
            reason: "fuzzTest",
            retryOnFailure: false,
        });
        return Promise.all([
            result.summarySubmitted,
            result.summaryOpBroadcasted,
            result.receivedSummaryAckOrNack,
        ]);
    }
    async submitSummary(options) {
        const handle = uuid();
        const summaryMessage = {
            handle,
            head: "",
            message: "",
            parents: [],
        };
        const clientSequenceNumber = ++this.deltaManager.clientSequenceNumber;
        const referenceSequenceNumber = this.deltaManager.lastSequenceNumber;
        const minimumSequenceNumber = this.factory.getMinSeq();
        const summarizeMessage = {
            type: MessageType.Summarize,
            clientSequenceNumber,
            referenceSequenceNumber,
            contents: summaryMessage,
        };
        this.deltaManager.outbound.push([summarizeMessage]);
        this.addPendingMessage(summarizeMessage.contents, summarizeMessage.metadata, summarizeMessage.clientSequenceNumber);
        this.factory.processAllMessages();
        this.scheduleAckNack(this.nackScheduled /* isNack */, handle, this.deltaManager.lastSequenceNumber);
        this.nackScheduled = false;
        const summaryStats = {
            ...mergeStats(),
            dataStoreCount: 1,
            summarizedDataStoreCount: 1,
            summaryNumber: 1,
        };
        return {
            stage: "submit",
            handle,
            clientSequenceNumber,
            referenceSequenceNumber,
            minimumSequenceNumber,
            submitOpDuration: 0,
            uploadDuration: 0,
            generateDuration: 0,
            summaryTree: {
                type: SummaryType.Tree,
                tree: {},
            },
            summaryStats,
        };
    }
    scheduleAckNack(isNack, handle, summarySequenceNumber) {
        // eslint-disable-next-line @typescript-eslint/no-floating-promises
        Promise.resolve().then(() => {
            const contents = {
                handle,
                summaryProposal: {
                    summarySequenceNumber,
                },
            };
            const summaryAckMessage = {
                type: isNack ? MessageType.SummaryNack : MessageType.SummaryAck,
                clientSequenceNumber: ++this.deltaManager.clientSequenceNumber,
                referenceSequenceNumber: this.deltaManager.lastSequenceNumber,
                contents,
            };
            this.deltaManager.outbound.push([summaryAckMessage]);
            this.addPendingMessage(summaryAckMessage.contents, undefined, summaryAckMessage.clientSequenceNumber);
            this.factory.processAllMessages();
        });
    }
    async refreshLatestSummaryAck(options) {
        // Do nothing
    }
    setConnectedState(value) {
        super.setConnectedState(value);
        this.connectedState.setConnectedState(value, this.clientId);
        this.summarizerClientElection.setClientId(this.clientId);
    }
    closeFn() {
        this.disposeFn();
    }
    disposed = false;
    disposeFn() {
        this.connected = false;
        this.disposed = true;
        this.summaryManager.dispose();
        this.summarizer.dispose();
        this.deltaManager.dispose();
    }
}
class MockSummarizerClientElection extends TypedEventEmitter {
    electedClientId;
    electedParentId;
    constructor(clientId) {
        super();
        this.setClientId(clientId);
    }
    setClientId(clientId) {
        this.electedClientId = clientId;
        this.electedParentId = clientId;
    }
}
class MockConnectedState extends TypedEventEmitter {
    logger;
    clientId;
    connected = false;
    constructor(logger, clientId) {
        super();
        this.logger = logger;
        this.clientId = clientId;
    }
    setConnectedState(connected, clientId) {
        if (this.connected === connected) {
            return;
        }
        this.clientId = clientId;
        raiseConnectedEvent(this.logger, this, connected, clientId);
    }
}
class MockThrottler {
    getDelay = () => 0;
    numAttempts = 3;
    delayWindowMs = 0;
    maxDelayMs = 0;
    delayFn = () => 0;
}
//# sourceMappingURL=summarizerFuzzMocks.js.map